/*
 * OptionsOracle
 * Copyright 2006-2012 SamoaSky (Shlomo Shachar & Oren Moshe)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

using System;
using System.Collections.Generic;
using System.Collections;
using System.Text;
 using OptionsOracle.Calc.Statistics;

using SharpCharts.Base.Data;

namespace OptionsOracle.Calc.Options
{
    public abstract class OptionPricingModel
    {
    
    	private const double MIN_VOLATILITY =  0.0;
        private const double MAX_VOLATILITY = 10.0;
        private const double VOLATILITY_ITERATIONS = 100;
        private const double VOLATILITY_ACCURACY = 0.001;
        
        protected static NormalDist nd = new NormalDist(0, 1.0);

       
        protected static void D(double p, double s, double r, double d, double v, double t, out double d1, out double d2)
        {
            d1 = (Math.Log(p / s) + (r - d + v * v / 2) * t) / (v * Math.Sqrt(t));
            d2 = d1 - (v * Math.Sqrt(t));
        }

		
        public virtual double Delta(OptionType type, double p, double s, double r, double d, double v, double t)
        {
            if (t <= 0) return 1;

            double d1, d2;
            D(p, s, r, d, v, t, out d1, out d2);

            if (type == OptionType.Call) return Math.Exp(-d * t) * nd.CDF(d1);
            else return -Math.Exp(-d * t) * nd.CDF(-d1);
        }
		
		
		       
        public virtual double Gamma(double p, double s, double r, double d, double v, double t)
        {
            if (t <= 0) return 0;

            double d1, d2;
            D(p, s, r, d, v, t, out d1, out d2);
            
            return Math.Exp(-d * t) * nd.PDF(d1) / (p * v * Math.Sqrt(t));
        }

	public virtual double Vega(double p, double s, double r, double d, double v, double t)
        {
            if (t <= 0) return 0;

            double d1, d2;
            D(p, s, r, d, v, t, out d1, out d2);

            return p * Math.Exp(-d * t) * nd.PDF(d1) * Math.Sqrt(t);
        }

        public virtual double Theta(OptionType type, double p, double s, double r, double d, double v, double t)
        {
            if (t <= 0) return 0;

            double d1, d2;
            D(p, s, r, d, v, t, out d1, out d2);

            if (type == OptionType.Call)
            {
                return -(p * Math.Exp(-d * t) * nd.PDF(d1) * v) / (2 * Math.Sqrt(t)) - r * s * Math.Exp(-r * t) * nd.CDF(d2) - d * p * Math.Exp(-d * t) * nd.CDF(d1);
            }
            else
            {
                return -(p * Math.Exp(-d * t) * nd.PDF(d1) * v) / (2 * Math.Sqrt(t)) + r * s * Math.Exp(-r * t) * nd.CDF(-d2) + d * p * Math.Exp(-d * t) * nd.CDF(-d1);
            }
        }
        
        public virtual double Rho(OptionType optionType, double p, double s, double r, double d, double v, double t)
        {
            if (t <= 0) return 0;

            double d1, d2;
            D(p, s, r, d, v, t, out d1, out d2);

            if (optionType == OptionType.Call)
            {
                return s * t * Math.Exp(-r * t) * nd.CDF(d2);
            }
            else
            {
                return s * t * Math.Exp(-r * t) * nd.CDF(-d2);
            }
        }
        
        
        public virtual double ImpliedVolatility(OptionType type, double o, double p, double s, double r, double d, double t)
        {
            // check incoming variables
            if (t <= 0 || double.IsNaN(o) || double.IsNaN(p) || double.IsNaN(s) || double.IsNaN(r) || double.IsNaN(t)) return double.NaN;

            int i;
            double min = MIN_VOLATILITY, max = MAX_VOLATILITY;
            double x, e, v = double.NaN, delta;

            for (i = 0; i < VOLATILITY_ITERATIONS; i++)
            {
                // calculate option price with guessed volatility
                v = (min + max) * 0.5;
                x = TheoreticalOptionPrice(type, p, s, r, d, v, t, out delta);

                // check result accuracy
                e = o - x;
                if (Math.Abs(e) < VOLATILITY_ACCURACY)
                {
                    if (v < 0.01) return double.NaN;
                    else return v;
                }

                // set
                if (e > 0) min = v;
                else max = v;
            }

            if (v < 0.01 || i == VOLATILITY_ITERATIONS) return double.NaN;
            else return v;
        }
        
        
        
        public abstract double TheoreticalOptionPrice(OptionType type, double p, double s, double r, double d, double v, double t, out double delta);

       
        //public abstract double TheoreticalOptionPrice(OptionType type, double p, double s, double r, double d, double v, double t);

        
    }
}
